#include <iostream>
#include <portable_io.h>
#include <cxxtls/csv.h>
#include <parse_date.h>
#include <string>
#include <vector>
#include <set>
#include <fstream>
#include <cxxtls/vectorVector.h>


using namespace std;
using namespace cxxtls;
using namespace vectorVector;


void printHelpAndDie();
void sortFile(istream &cin, Sorter const &keys, bool passthrough);

int main(int argc, char **argv)
{
    ++argv;

    bool passthrough = false;

    Sorter keys;

    while(*argv)
    {
        string curArg = *argv++;

        bool moreParms = (*argv != 0);

        if(curArg == "-f")
        {
           passthrough = true;
           break;
        }
        else
        if(curArg == "-h" || curArg == "--help")
        {
            printHelpAndDie();
        }
        else
        if((curArg == "-s" || curArg == "-S") && moreParms)
        {
           bool reverse = curArg[1] == 'S';

           curArg = *argv++;


           int fieldNo = -1;

           sscanf(curArg.c_str(), "%d", &fieldNo);

           if(fieldNo < 1)
           {
                cerr << "sortcsv:  error:  invalid -s index:  " << curArg << endl;
                exit(1);
           }

           --fieldNo; // switch from 1 base to zero based numbers

           keys.push_back( SortKey(fieldNo, SortKey::SK_string, reverse));

        }
        else
        if((curArg == "-N" || curArg == "-n") && moreParms)
        {
           bool reverse = curArg[1] == 'N';

           curArg = *argv++;

           int fieldNo = -1;

           sscanf(curArg.c_str(), "%d", &fieldNo);

           if(fieldNo < 1)
           {
                cerr << "sortcsv:  error:  invalid -n index value:  " << curArg << endl;
                exit(1);
           }

           --fieldNo;  // switch from 1 base to zero based numbers


           keys.push_back( SortKey(fieldNo, SortKey::SK_number, reverse));

        }
        else
        if((curArg == "-d" || curArg == "-D") && moreParms)
        {
           bool reverse = curArg[1] == 'D';

           curArg = *argv++;

           int fieldNo = -1;

           sscanf(curArg.c_str(), "%d", &fieldNo);

           if(fieldNo < 1)
           {
                cerr << "sortcsv:  error:  invalid -d  index value:  " << curArg << endl;
                exit(1);
           }

           --fieldNo;  // switch from 1 base to zero based numbers

           keys.push_back( SortKey(fieldNo, SortKey::SK_date, reverse));

        }
        else
        if(curArg[0] != '-')
        {
           --argv; // push the most recent parm back into the input stream
           break;
        }
        else
        if(curArg == "-")
        {
          --argv; // push - back into the input stream.
          break;
        }
        else
        {
            cerr << "sortcsv:  error:  invalid option or missing sub-option:  " << curArg << endl;
            exit(1);
        }

    }

    if(passthrough)
    {
       if(!keys.empty())
          printHelpAndDie();
    }
    else
    if(keys.empty())
       printHelpAndDie();


    if(*argv != 0)
    {
       string curArg = *argv++;

       if(curArg != "-")
       {
           ifstream file(curArg.c_str());

           if(!file.good())
           {
              cerr << "sortcsv:  error unable to open file '" << curArg << "'" << endl;
              exit(1);
           }

           sortFile(file, keys, passthrough);

           exit(0);
       }
        
    }

    sortFile(cin, keys, passthrough);

    exit(0);


}

void sortFile(istream &cin, Sorter const &keys, bool passthrough)
{
    string line;

    if(passthrough)
    {
        // passthrough mode, or 'filter mode', is just a simple
        // copy operation.  We parse the input data and then
        // re-encode it as a string then print that.

        while(std::getline(cin, line))
        {
            vector<string>  columns;
        
            csv::split(line, columns);

            string tmp;

            csv::format(columns, tmp);
            
            cout << tmp << endl;
        
        }

        return;
    }


    multiset<vector<string>, Sorter> output(keys);

    while(std::getline(cin, line))
    {
        vector<string>  columns;

        csv::split(line, columns);

        while(columns.size() <= keys.minIndex_ )
            columns.push_back(string());

        output.insert(columns);

    }

    // output is by default sorted...

    multiset<vector<string>, Sorter>::const_iterator begin = output.begin(), end = output.end();

    while(begin != end)
    {
        string tmp;

        csv::format(*begin++, tmp);

        cout << tmp << endl;
    }
}


void printHelpAndDie()
{
    cerr << "sortcsv:  a program that sorts comma separated values files\n"
            "          based on user defined key definitions.  A 'key' is\n"
            "          is a definition of a column number and the type of\n"
            "          sorting to be done on that column.   The types of\n"
            "          sorting are: \n"
            "\n"
            "             -s     string comparison\n"
            "             -n     numeric (ie floating point) comparison\n"
            "             -d     date comparison\n"
            "             -S     reverse string comparison\n"
            "             -N     reverse numeric (ie floating point) comparison\n"
            "             -D     reverse date comparison\n"
            "\n"
            "Options:\n"
            "          sortcsv [ (-s | -n | -d | -S | -N | -D) fieldNo  ] [filename] \n"
            "          sortcsv -f [filename] \n"
            "\n"
            "The first field is number 1\n"
            "\n"
            "In abscence of a named file, stdin is read.  Output always\n"
            "goes to stdout.  If the filename is - then stdin is read.\n"
            "\n"
            "The -f option means to read the csv stream but not to\n"
            "sort it.  This option is useful for eliminating tabs\n"
            "and extraneous spaces.\n"
            "\n"
            "Date Formats:\n\n"
            "    24/12/02\n"
            "    02/02/99\n"
            "    15/07/2002\n"
            "    02/01/02 10:45\n"
            "    01/03/02 10:45:01\n"
            "    Mon, Jan 13 1980\n"
            "    Tues, Dec 7 1941 10:45 -05:00\n"
            "    Feb 14 1918 10:45\n"
            "    Feb 14 2002 10:45:03\n"
            "    Sat, October 13, 2003 09:15:59 -05:00\n"
            "\n"
         ;

     exit(1);
}
