#ifndef portable_strstream_included
#define portable_strstream_included

///@file portable_strstream.h
///   Defines an interface to class strtream::strstream that
///   is portable across linux, windows, and major unixes.
///


#ifndef LINUX
#include <strstream>
#else

#include <iterator>
#include <strstream>

#if !defined(__GNUC__) || __GNUC__  < 3
namespace std
{
  using namespace std;


  template<class charT>
  class  istreambuf_iterator
  : public input_iterator<charT, int>
    ///  fix for missing istreambuf_iterator in g++ 2.95
  {

  public:
    typedef charT                          char_type;
    typedef int                            int_type;

    typedef std::streambuf                 streambuf_type;
    typedef std::istream                   istream_type;

    class proxy {
      char_type                       __keep;
      streambuf_type                  *__sbuf;

      proxy(char_type c, streambuf_type *sbuf)
	: __keep(c), __sbuf(sbuf)
      { ; }

    public:

      char_type operator*()
      { return __keep; }

      friend class istreambuf_iterator<charT>;
    };

  public:

    istreambuf_iterator()
    : __sbuf(0)
    { __failed_flag = true; }
    istreambuf_iterator(istream_type& s)
    : __sbuf(s.rdbuf())
    {
      if ( s.rdbuf() ) __failed_flag = false;
      else
	__failed_flag = true;
    }
    istreambuf_iterator(streambuf_type *s)
    : __sbuf(s)
    {
      if ( s ) __failed_flag = false;
      else
	__failed_flag = true;
    }
    istreambuf_iterator(const proxy& p)
    : __sbuf(p.__sbuf)
    { ; }
    inline char_type operator*();
    inline istreambuf_iterator<charT>& operator++();
    inline proxy operator++(int);
    inline bool equal(istreambuf_iterator<charT>& b);

    bool failed( ) const
    { return __failed_flag; }

  protected:

  private:
    streambuf_type     *__sbuf;
    bool                __failed_flag;
  };

  template<class charT>
  inline typename istreambuf_iterator<charT>::char_type
  istreambuf_iterator<charT>::operator*()
  {
    int_type c;

    if ( __sbuf && !__failed_flag )
    {
      c= __sbuf->sgetc();
      if ( c == EOF )
      {
	__sbuf = 0;
	__failed_flag = true;
      }
    }
    else return EOF;

    return c;
  }

  template<class charT>
  inline istreambuf_iterator<charT>&
  istreambuf_iterator<charT>::operator++()
  {
    if (__sbuf && !__failed_flag )
    {
      __sbuf->sbumpc();
      if ( __sbuf->sgetc() == EOF )
      {
	__sbuf = 0;
	__failed_flag = true;
      }
    }
    return *this;
  }


  template<class charT>
  inline typename istreambuf_iterator<charT>::proxy
  istreambuf_iterator<charT>::operator++(int)
  {

    if (__sbuf && !__failed_flag )
    {
      proxy     prev(__sbuf->sgetc(), __sbuf);
      __sbuf->sbumpc();
      if ( __sbuf->sgetc() == EOF )
      {
	__sbuf = 0;
	__failed_flag = true;
      }
      return prev;
    }

    charT     c=EOF;
    return proxy(c, __sbuf);
  }


  template<class charT>
  inline bool
  istreambuf_iterator<charT>::
  equal(istreambuf_iterator<charT>& b)
  {
    if( ((__sbuf ==0) && (b.__sbuf==0)) || ((__sbuf !=0) && (b.__sbuf !=0)) )
      return true;
    else
      return false;
  }


  template<class charT>
  inline bool  operator==(istreambuf_iterator<charT>& a,
					      istreambuf_iterator<charT>& b)
  {
    return a.equal(b);
  }


  template<class charT>
  inline bool  operator!=(istreambuf_iterator<charT>& a,
					      istreambuf_iterator<charT>& b)
  {
    return !(a.equal(b));
  }

template< class charT >
class ostreambuf_iterator
: public output_iterator
{

public:
  //
  // Types:
  //
  typedef charT                          char_type;
  typedef std::streambuf                 streambuf_type;
  typedef std::ostream                   ostream_type;

  ostreambuf_iterator(ostream_type& s)
  : __sbuf(s.rdbuf())
  {
    if ( s.rdbuf() ) __failed_flag = false;
    else
      __failed_flag = true;
  }

  ostreambuf_iterator(streambuf_type *s)
  : __sbuf(s)
  {
    if ( s ) __failed_flag = false;
    else
      __failed_flag = true;
  }

  ostreambuf_iterator< charT >& operator*()
  { return *this; }
  ostreambuf_iterator< charT >& operator++()
  { return *this; }
  ostreambuf_iterator< charT > operator++(int)
  { return *this; }

  ostreambuf_iterator< charT >& operator=(charT c)
  {
    if ( !__failed_flag )
    {
      if ( __sbuf->sputc(c)  == EOF )
	__failed_flag=true;
    }
    return *this;
  }

  bool failed( ) const
  { return __failed_flag; }

protected:

private:
  streambuf_type        *__sbuf;
  bool                   __failed_flag;
};


};

#endif


#endif
#endif
