#ifdef _MSC_VER
#include <stdio.h>
#include <stdlib.h>

#include <WindowsUnixIo.h>

#include <cxxtls/commandLineOptions.h>
#include <cxxtls/exception.h>
#include <cxxtls/lexical_cast.h>
#include <cxxtls/ThisApplication.h>



#include <sstream>

using namespace std;

void tailFile(WindowsFileHandle &sout, 
              bool               follow,
              size_t             linesToDisplayAtFirst, 
              std::string const &filename
             );  


DECLARE_CXXTLS_EXCEPTION(TAIL);

static void displayInitialLineGroup( WindowsFileHandle &fd, 
                                     WindowsFileHandle &sout, 
                                     size_t            &oldSize, 
                                     size_t linesToDisplayAtFirst 
                                   );

void tailFile(WindowsFileHandle &sout, bool follow, size_t linesToDisplayAtFirst, std::string const &filename)  
    // performs the actual work.  Throws std::exception on error
{
    WindowsFileHandle fd;

    if(!fd.open(filename))
    {
        throw TAIL("Can't Open Input File", filename);
    }

    // destructor for fd will close the handle...

    try
    {

        WindowsFileHandle::FileStatus statbuf;

        if(!fd.getStatus(statbuf))
        {
            throw TAIL("Unable to get size of File", filename);
        }
    
        size_t oldSize = statbuf.size_;
    
        displayInitialLineGroup( fd, sout, oldSize, linesToDisplayAtFirst );  // levaes file at new end, with oldSize updated
    
        if(!follow)
        {
           return; // displaying the leading data is all that is desired.
        }
    
    
        for(;;)
        {
            fd.close();

            for(int i = 0; i < 100; ++i)  
            {
                if(cxxtls::shuttingDown())
                   return;

                Sleep(10);  // sleep in 1/100 of a second increments to speed up ^C if it is used
            }
            if(!fd.open(filename))
            {
                throw TAIL("Can't RE-Open Input File", filename);
            }

            if(!fd.getStatus(statbuf))
            {
                throw TAIL("Unable to get size of File", filename);
            }
    
            long newSize = statbuf.size_;
    
            if(newSize < oldSize)
            {
               oldSize = newSize;
            }

            fd.seek(oldSize);


            if(oldSize != newSize)
            {
                 char buffer[32768];
    
                 int bytes;
    
                 while(0 < (bytes = fd.read(buffer, sizeof(buffer)) ) )
                    sout.write(buffer, bytes);
    
                 if(!fd.getStatus(statbuf))
                 {
                     throw TAIL("Unable to get size of File", filename);
                 }
    
                oldSize = statbuf.size_;
    
            }
    
        }

    }
    catch(...)
    {
        throw;  // rethrow after closing the input file
    }

}


// scan the file backwards from oldSize.  Update linesToDisplayAtFirst and firstTimeThrough
// look for the beginning of the line defined by linesToDisplayAtFirst.  When it is found,
// print from there until end of file.

static bool readBackwardsFrom(WindowsFileHandle &fd, 
                              size_t& oldSize,
                              size_t& linesToDisplayAtFirst,
                              bool  & firstTimeThrough
                             )
{
    char buffer[32768];


    while(linesToDisplayAtFirst)
    {
        
        if(oldSize > sizeof(buffer))
        {
           oldSize -= sizeof(buffer);
        }
        else
        {
           oldSize = 0;
        }

        if(!fd.seek(oldSize))
        {
           return false;  // can't seek to proper position
        }

        int readBytes = fd.read(buffer, sizeof(buffer));

        if(readBytes < 1)
           return false;

        char const *scan = &buffer[readBytes];
        char const *begin= &buffer[0];

        while(scan > begin)
        {
           --scan;

           if(*scan != '\n')
           {
               firstTimeThrough = false;
               continue;
           }

           if(firstTimeThrough)
           {
              firstTimeThrough=false;
              continue;
           }

           --linesToDisplayAtFirst;

           if(linesToDisplayAtFirst == 0)
           { 
              // we have found the thing we are looking for

              ++scan;  // scan points to first byte of last row

              oldSize +=  (scan-begin);

              return true;  // found it

           }

        }

        // at this point, scan should == begin  and the current buffer
        // did not have the data we are looking for, but we may have
        // updated linesTodisplayAtFirst

        if(oldSize == 0)
        {
           // we have read back to the beginning of the file

           return true;  // pretend this is a success
        }

    }


    return true; // found it.

}



static void displayInitialLineGroup( WindowsFileHandle &fd, 
                                     WindowsFileHandle &sout,  // standard out
                                     size_t& oldSize, 
                                     size_t linesToDisplayAtFirst 
                                   )
{
    bool firstTimeThrough = true;

    if(linesToDisplayAtFirst == 0)
       return;

    size_t originalSize = oldSize;

    bool foundIt = readBackwardsFrom(fd, oldSize, linesToDisplayAtFirst, firstTimeThrough);

    if(!foundIt)
    {
       fd.seek(originalSize);

       return;   
    }

    // oldSize has been updated

    if(!fd.seek(oldSize))
    {

       throw TAIL("Unable to seek to file location", "someplace");

    }

    char buffer[32768];
    int  bytesRead;

    for(;;)
    {
        bytesRead = fd.read(buffer, sizeof(buffer));

        if(bytesRead < 1)
         break;

        sout.write(buffer, bytesRead);

        oldSize += bytesRead;
    }


}
#else
namespace {
    int pleaseignorethisvariable=0;
    }
#endif
